from copy import deepcopy
from math import log
hg19 = [249250621, 243199373, 198022430, 191154276, 180915260, 171115067, 159138663, 146364022, 141213431, 135534747, 135006516, 133851895, 115169878, 107349540, 102531392, 90354753, 81195210, 78077248, 59128983, 63025520, 48129895, 51304566, 155270560, 59373566]
hg18 = [247249719, 242951149, 199501827, 191273063, 180857866, 170899992, 158821424, 146274826, 140273252, 135374737, 134452384, 132349534, 114142980, 106368585, 100338915, 88827254, 78774742, 76117153, 63811651, 62435964, 46944323, 49691432, 154913754, 57772954]
mm9 = [197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255, 121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430, 166650296, 15902555]

def genome_version(gv):

    """Read Genome Version, currently supporting hg18, hg19, mm9"""    
    chromosomes = []
    if gv == 'hg18':
        chromosomes = hg18
    elif gv == 'hg19':
        chromosomes = hg19
    elif gv == 'mm9':
        chromosomes = mm9
    return chromosomes

def read_hic_file(infile, filter):

    intra_hic = [[] for i in xrange(25)]
    filter_intra_hic = [[] for i in xrange(25)]
    inter_hic = []
    inf = open(infile, 'r')
    chrom = []
    #Read input line by line
    for line in inf:
        a = line.split()
        a[2], a[6] = int(a[2]), int(a[6])
        if a[1] == a[5] and a[1] != '0':
            if int(a[1]) not in chrom:
                chrom.append(int(a[1]))
            if a[2] < a[6]:
                None
            else:
                a[2], a[6] = a[6], a[2]
            intra_hic[int(a[1])-1].append([a[2], a[6]])
            if abs(a[2] - a[6]) >= filter:  #Generate the filtered hi-c file
                filter_intra_hic[int(a[1]) - 1].append([a[2], a[6]])
        #if a[1] != a[5]:
            #inter_hic.append([int(a[1]), int(a[2]), int(a[5]), int(a[6])])
    inf.close()
    chrom.sort()
    chr_la = chrom[-1]
    intra_hic = intra_hic[:chr_la]  #Only considering the chromosomes in Hi-C file
    filter_intra_hic = filter_intra_hic[:chr_la]
    for i in xrange(len(intra_hic)):
        intra_hic[i].sort()
        filter_intra_hic[i].sort()
    
    return chrom, intra_hic, filter_intra_hic, inter_hic
    
def interaction_ratio(hic, filter_hic, resolution, c, chromosomes):

    temp = []
    filter_vect = []
    filter_abs_vect = []
    #Transit the tag position into XXX resolution
    for i in xrange(len(filter_hic)):
        filter_hic[i][0] = float(filter_hic[i][0])/resolution
        filter_hic[i][1] = float(filter_hic[i][1])/resolution
        u1, u2 = int(filter_hic[i][0]), int(filter_hic[i][1])
        temp.append([u1, u2, 1])
    temp.sort()
    block = chromosomes[c]/resolution
    filter_ave_rat = [0 for i in xrange(block+1)]
    i = 0
    #Eliminate the repeated interactions with respect to the current resolution
    while i < len(temp) - 1:
        a = deepcopy(temp[i])
        filter_vect.append(a)
        for j in xrange(i + 1, len(temp)):
            if temp[j] == temp[i]:
                filter_vect[-1][2] += 1
            elif temp[j] != temp[i]:
                i = j
                break
        if j == len(temp) - 1:
            break
    del(temp)
    filter_vect.sort()
    filter_abs_vect = deepcopy(filter_vect)
    #Calculate the average interaction value with the same distance
    for v in filter_vect:
        filter_ave_rat[v[1] - v[0]] += v[2]
    for i in xrange(len(filter_ave_rat)):
        filter_ave_rat[i] = float(filter_ave_rat[i]) / (len(filter_ave_rat) - i)
    #Normalize the interaction values
    for i in xrange(len(filter_vect)):
        m = filter_ave_rat[filter_vect[i][1] - filter_vect[i][0]]
        if m == 0:
            filter_vect[i][2] = 0
        else:
            filter_vect[i][2] = float(filter_vect[i][2]) / m
    
    #Same as above, except considering the non-filtered interactions
    vect = []
    abs_vect = []  
    ave_rat, temp = [0 for i in xrange(block+1)], []
    for i in xrange(len(hic)):
        hic[i][0] = float(hic[i][0])/resolution
        hic[i][1] = float(hic[i][1])/resolution
        u1, u2 = int(hic[i][0]), int(hic[i][1])
        temp.append([u1, u2, 1])
    temp.sort()
    i = 0
    while i < len(temp) - 1:
        a = deepcopy(temp[i])
        vect.append(a)
        for j in xrange(i + 1, len(temp)):
            if temp[j] == temp[i]:
                vect[-1][2] += 1
            elif temp[j] != temp[i]:
                i = j
                break
        if j == len(temp) - 1:
            break
    del(temp)
    vect.sort()
    abs_vect = deepcopy(vect)
    for v in vect:
        ave_rat[v[1] - v[0]] += v[2]
    for i in xrange(len(ave_rat)):
        ave_rat[i] = float(ave_rat[i]) / (len(ave_rat) - i)
    for i in xrange(len(vect)):
        m = ave_rat[vect[i][1] - vect[i][0]]
        if m == 0:
            vect[i][2] = 0
        else:
            vect[i][2] = float(vect[i][2]) / m
    return vect, abs_vect, filter_vect, filter_abs_vect

def domain_ratio(vect, distal, resolution, c, chromosomes):

    offset = distal / resolution
    block = chromosomes[c] / resolution
    ldr = []
    #Compute the sum of closed interactions and long interactions
    dom_ratio = [[0, 0, 0] for i in xrange(block + 1)]
    for vc in vect:
        a, b = vc[0], vc[1]
        if 0 < abs(a - b) <= offset/2:
            dom_ratio[a][0] += vc[2]
            dom_ratio[b][0] += vc[2]
        elif abs(a - b) == 0:
            dom_ratio[a][0] += vc[2]
        elif abs(a - b) > offset/2:
            dom_ratio[a][1] += vc[2]
            dom_ratio[b][1] += vc[2]
    #Compute the linkage ratios
    for i in xrange(len(dom_ratio)):
        m, p = i - offset/2, i + offset/2
        ic, il = 0, 0
        l = len(dom_ratio)
        if m >= 0 and p < l:
            ic = p + 1 - m
            il = m + l - p - 1
        elif m >= 0 and p >= l:
            ic = l - m
            il = m
        elif m < 0 and p < l:
            ic = p + 1
            il = l - 1 - p
        if dom_ratio[i][0] != 0:
            dom_ratio[i][0], dom_ratio[i][1] = float(dom_ratio[i][0])/ic, float(dom_ratio[i][1])/il
            dom_ratio[i][2] = dom_ratio[i][1] / dom_ratio[i][0]
        if dom_ratio[i][2] != 0:
            ldr.append(-log(dom_ratio[i][2], 2))
        else:
            ldr.append(0)
    del(dom_ratio)
    return ldr

def select_chrom(ldr, vect, abs_vect):

    #Select the different types of interactions
    eu_eu, he_he, eu_he = [], [], []
    abs_eu_eu, abs_he_he, abs_eu_he = [], [], []
    euregion, heteroregion = [], []
    for i in xrange(len(ldr)):
        l = ldr[i]
        if l > 0:
            euregion.append(i)
        elif l < 0:
            heteroregion.append(i)
    for v in vect:
        a, b, i, j, l, m = v[0], v[1], 0, 0, 0, 0
        for r in euregion:
            if a == r:
                i = 1
                break
        for r in euregion:
            if b == r:
                j = 1
                break
        for r in heteroregion:
            if a == r:
                l = 1
                break
        for r in heteroregion:
            if b == r:
                m = 1
                break
        if i == 1 and j == 1:
            eu_eu.append(v)
        elif l == 1 and m == 1:
            he_he.append(v)
        else:
            eu_he.append(v)
    for v in abs_vect:
        a, b, i, j, l, m = v[0], v[1], 0, 0, 0, 0
        for r in euregion:
            if a == r:
                i = 1
                break
        for r in euregion:
            if b == r:
                j = 1
                break
        for r in heteroregion:
            if a == r:
                l = 1
                break
        for r in heteroregion:
            if b == r:
                m = 1
                break
        if i == 1 and j == 1:
            abs_eu_eu.append(v)
        elif l == 1 and m == 1:
            abs_he_he.append(v)
        else:
            abs_eu_he.append(v)
    return eu_eu, he_he, eu_he, abs_eu_eu, abs_he_he, abs_eu_he
